#![recursion_limit = "256"]
#[macro_use]
extern crate clap;
#[macro_use]
extern crate log;
use bytes::buf::Buf;
use futures::future::Either;
use futures::sink::Sink;
use futures::stream::Stream;
use futures::{Async, Future, Poll};
use rand::distributions::Alphanumeric;
use rand::{thread_rng, Rng};
use std::borrow::Cow;
use std::io::Write;
use std::process::Command;
use std::sync::Arc;
use warp::http::header::{HeaderMap, HeaderValue};
use warp::Filter;

mod notify;
use notify::Notify;

include!(concat!(env!("OUT_DIR"), "/static_files.rs"));
const DEFAULT_PORT_STR: &'static str = "4000";
fn main() {
    env_logger::init();

    let app = clap::App::new("Codenano server")
        .version(crate_version!())
        .author("P.E. Meunier")
        .arg(if cfg!(feature = "docker") {
            clap::Arg::with_name("file")
                .help("Serve a file from the filesystem")
                .takes_value(true)
        } else {
            clap::Arg::with_name("file")
                .help("Serve a file from the filesystem")
                .required_unless("designs")
                .takes_value(true)
        })
        .arg(
            clap::Arg::with_name("port")
                .short("p")
                .validator(|p| {
                    p.parse::<u16>()
                        .map(|_| ())
                        .map_err(|_| "The port must be a number".to_string())
                })
                .help("Port on which to run the server")
                .takes_value(true),
        );
    let app = if !cfg!(feature = "docker") {
        app.arg(
            clap::Arg::with_name("designs")
                .long("designs")
                .help("Compilation directory")
                .required_unless("file")
                .takes_value(true),
        )
    } else {
        app
    };
    let matches = app.get_matches();

    let designs = Arc::new(
        matches
            .value_of("designs")
            .map(|x| x.to_string())
            .unwrap_or(String::new()),
    );
    let designs_ = designs.clone();
    let mut headers = HeaderMap::new();
    headers.insert("Content-Type", HeaderValue::from_static("application/json"));
    headers.insert(
        "Content-Disposition",
        HeaderValue::from_static("attachement; filename = \"cadnano_file.json\""),
    );

    /*
    let export_cadnano = warp::post2()
    .and(warp::path("export-cadnano.json"))
    .and(warp::body::concat())
    .and_then(move |body| {
        if cfg!(feature = "docker") {
            let name: String = std::iter::once('z')
                .chain(thread_rng().sample_iter(&Alphanumeric).take(30))
                .collect();
            export_cadnano(body, name)
        } else {
            export_cadnano(body, designs.clone().to_string())
        }
    })
    .with(warp::reply::with::headers(headers));
    */
    let export_cadnano = warp::path("export_cadnano")
        .and(warp::body::concat())
        .map(move |body| export_json(body))
        .with(warp::reply::with::headers(headers));

    let mut headers_json = HeaderMap::new();
    headers_json.insert("Content-Type", HeaderValue::from_static("application/json"));
    headers_json.insert(
        "Content-Disposition",
        HeaderValue::from_static("attachement; filename = \"modified.json\""),
    );

    let export_json = warp::path("export_json")
        .and(warp::body::concat())
        .map(move |body| export_json(body))
        .with(warp::reply::with::headers(headers_json));

    let port: u16 = matches
        .value_of("port")
        .unwrap_or(DEFAULT_PORT_STR)
        .parse()
        .expect("Could not parse port");
    let (ws_tx, ws_rx) = tokio::sync::watch::channel(String::new());
    let (update_tx, update_rx) = tokio::sync::mpsc::unbounded_channel();
    let thr = || {
        std::thread::spawn(move || {
            // Thread that checks for updates and sends it back to the websocket.
            tokio::run(
                Notify::new(None, update_rx)
                    .fold(ws_tx, |ws_tx, msg| {
                        let mut body = String::new();
                        if let Some(msg) = msg {
                            if let Ok(mut f) = std::fs::File::open(&msg) {
                                use std::io::Read;
                                f.read_to_string(&mut body).unwrap();
                                return Either::A(
                                    ws_tx.send(body).map_err(|e| error!("ws tx err: {:?}", e)),
                                );
                            }
                        }
                        Either::B(futures::finished(ws_tx))
                    })
                    .map(|_| ()),
            )
        })
    };
    let update = update(update_tx.clone());
    if let Some(path) = matches.value_of("file") {
        let path = path.to_string();
        update_tx.clone().try_send(path.clone()).unwrap();
        let path_ = Arc::new(path.to_string());
        let websocket = warp::path("websocket")
            .and(warp::ws2())
            .map(move |ws: warp::ws::Ws2| websocket(ws, ws_rx.clone()));
        let run = warp::path("run").and(warp::post2().and(warp::fs::file(path)));
        let routes = run
            .or(export_cadnano)
            .or(websocket)
            .or(warp::path::end().and(index_ws()))
            .or(update)
            .or(static_files())
            .or(export_json);

        if let Ok((_, s)) = warp::serve(routes).try_bind_ephemeral(([127, 0, 0, 1], port)) {
            let thr = thr();
            hyper::rt::run(s);
            thr.join().unwrap();
        } else {
            let can = std::fs::canonicalize(path_.as_ref()).unwrap();
            let can: String = can.to_string_lossy().into();
            let req = hyper::Request::builder()
                .method("POST")
                .uri(format!("http://localhost:{}/update", port))
                .body(hyper::Body::from(can))
                .unwrap();
            let mut rt = tokio::runtime::current_thread::Runtime::new().unwrap();
            rt.block_on(
                hyper::Client::new()
                    .request(req)
                    .map(|_| ())
                    .map_err(|e| error!("Error {:?}", e)),
            )
            .unwrap()
        }
    } else {
        let thr = thr();
        let run = warp::post2()
            .and(warp::path("run"))
            .and(warp::body::concat())
            .and_then(move |body| {
                if cfg!(feature = "docker") {
                    let name: String = std::iter::once('z')
                        .chain(thread_rng().sample_iter(&Alphanumeric).take(30))
                        .collect();
                    run(body, name)
                } else {
                    run(body, designs_.clone().to_string())
                }
            });
        let routes = run
            // .or(export_cadnano)
            .or(warp::path::end().and(index()))
            .or(update)
            .or(static_files());
        warp::serve(routes).run(([127, 0, 0, 1], port));
        thr.join().unwrap();
    }
}

#[cfg(not(feature = "docker"))]
fn update(
    update_tx_: tokio::sync::mpsc::UnboundedSender<String>,
) -> impl Filter<Extract = (String,), Error = warp::Rejection> + Clone {
    warp::post2()
        .and(warp::path("update"))
        .and(warp::path::end())
        .and(warp::body::concat())
        .map(move |body: warp::filters::body::FullBody| {
            let body: Vec<u8> = body.collect();
            if let Ok(body) = String::from_utf8(body) {
                if let Err(e) = update_tx_.clone().try_send(body) {
                    format!("error {:?}", e)
                } else {
                    "Ok".to_string()
                }
            } else {
                "Non-UTF8 path".to_string()
            }
        })
}

#[cfg(feature = "docker")]
fn update(
    update_tx_: tokio::sync::mpsc::UnboundedSender<String>,
) -> impl Filter<Extract = (String,), Error = warp::Rejection> + Clone {
    warp::post2()
        .and(warp::path("update"))
        .and_then(|_| Err(warp::reject::not_found()))
}

fn websocket(
    ws: warp::ws::Ws2,
    rx: tokio::sync::watch::Receiver<String>,
) -> impl warp::reply::Reply {
    ws.on_upgrade(move |ws| {
        ws.send(warp::filters::ws::Message::text(String::new()))
            .map_err(|e| error!("err {:?}", e))
            .and_then(move |ws| {
                rx.map_err(|e| error!("err {:?}", e))
                    .fold(ws, move |ws, msg| {
                        if !msg.is_empty() {
                            Either::A(
                                ws.send(warp::filters::ws::Message::text(msg))
                                    .map_err(|e| error!("err {:?}", e)),
                            )
                        } else {
                            Either::B(futures::finished(ws))
                        }
                    })
                    .map(|_| ())
            })
    })
}

/*
#[cfg(target_os = "linux")]
fn websocket(ws: warp::ws::Ws2, path: Arc<String>) -> impl warp::reply::Reply {
    use inotify::{Inotify, WatchMask};

    let mut inotify = Inotify::init().expect("Error while initializing inotify instance");

    // Watch for modify and close events.
    inotify
        .add_watch(&path.as_ref(), WatchMask::CLOSE_WRITE)
        .expect("Failed to add file watch");

    // Read events that were added with `add_watch` above.
    ws.on_upgrade(move |ws| {
        let init = if let Ok(mut f) = std::fs::File::open(&path.as_ref()) {
            let mut body = String::new();
            use std::io::Read;
            f.read_to_string(&mut body).unwrap();
            Either::A(
                ws.send(warp::filters::ws::Message::text(body))
                    .map_err(|_| ()),
            )
        } else {
            Either::B(futures::finished(ws))
        };

        init.and_then(move |ws| {
            inotify
                .event_stream(vec![0; 1024])
                .map_err(|_| ())
                .fold(ws, move |ws, _event| {
                    let mut f = std::fs::File::open(&path.as_ref()).unwrap();
                    let mut body = String::new();
                    use std::io::Read;
                    f.read_to_string(&mut body).unwrap();
                    ws.send(warp::filters::ws::Message::text(body))
                        .map_err(|_| ())
                })
                .map(|_| ())
        })
    })
}
 */

fn run<P: AsRef<str> + 'static>(
    body: warp::filters::body::FullBody,
    name: P,
) -> impl Future<Item = String, Error = warp::reject::Rejection> {
    let body: Vec<_> = body.collect();
    let mut f = if cfg!(feature = "docker") {
        std::fs::File::create(name.as_ref()).unwrap()
    } else {
        let path = std::path::Path::new(name.as_ref())
            .join("src")
            .join("main.rs");
        std::fs::create_dir_all(path.parent().unwrap()).unwrap();
        debug!("path = {:?}", path);
        std::fs::File::create(&path).unwrap()
    };
    f.write_all(&body).unwrap();
    (RunFuture { file: Some(name) })
        .map(|x| serde_json::to_string(&x).unwrap())
        .map_err(|e| warp::reject::custom(e))
}

pub(crate) struct RunFuture<P: AsRef<str> + 'static> {
    file: Option<P>,
}

#[cfg(not(feature = "docker"))]
fn compile_and_run(name: &str) -> Result<codenano::Res, failure::Error> {
    println!("compile_and_run {:?}", name);
    let process = Command::new("cargo")
        .arg("run")
        .arg("-q")
        .current_dir(name)
        .env("RUST_BACKTRACE", "1")
        .output()?;
    println!("{}", std::str::from_utf8(&process.stderr).unwrap());
    debug!("{}", std::str::from_utf8(&process.stdout).unwrap());
    let out = if process.status.success() {
        String::from_utf8(process.stdout)?
    } else {
        debug!("{}", std::str::from_utf8(&process.stderr).unwrap());
        "".to_string()
    };
    Ok(codenano::Res {
        out,
        err: String::from_utf8_lossy(&process.stderr).into_owned(),
    })
}

#[cfg(feature = "docker")]
fn compile_and_run(name: &str) -> Result<codenano::Res, failure::Error> {
    println!("docker compile_and_run {:?}", name);
    Command::new("docker")
        .arg("run")
        .arg("--name")
        .arg(&name)
        .arg("--memory=200m")
        .arg("--cpus=1")
        .arg("--network")
        .arg("none")
        .arg("-td")
        .arg("codenano")
        .status()?;
    Command::new("docker")
        .arg("cp")
        .arg(&name)
        .arg(&format!("{}:appuser/src/main.rs", name))
        .status()?;

    let process = Command::new("docker")
        .arg("exec")
        .arg(&name)
        .arg("timeout")
        .arg("5s")
        .arg("sh")
        .arg("-c")
        .arg("cd appuser && RUST_BACKTRACE=1 REMOTE=1 cargo run -q")
        .output()?;
    debug!("{}", std::str::from_utf8(&process.stderr).unwrap());
    debug!("{}", std::str::from_utf8(&process.stdout).unwrap());
    let out = if process.status.success() {
        String::from_utf8(process.stdout)?
    } else {
        debug!("{}", std::str::from_utf8(&process.stderr).unwrap());
        "".to_string()
    };
    Command::new("docker").arg("kill").arg(&name).status()?;
    Command::new("docker").arg("rm").arg(&name).status()?;
    Ok(codenano::Res {
        out,
        err: String::from_utf8_lossy(&process.stderr).into_owned(),
    })
}

impl<P: AsRef<str> + 'static> Future for RunFuture<P> {
    type Item = codenano::Res;
    type Error = failure::Error;
    fn poll(&mut self) -> Poll<Self::Item, Self::Error> {
        match tokio_threadpool::blocking(|| {
            let name = self.file.take().unwrap();
            compile_and_run(name.as_ref())
        }) {
            Ok(Async::NotReady) => Ok(Async::NotReady),
            Ok(Async::Ready(x)) => Ok(Async::Ready(x?)),
            Err(e) => Err(e.into()),
        }
    }
}

/*
fn export_cadnano(
    body_: warp::filters::body::FullBody,
    name: String,
) -> impl Future<Item = String, Error = warp::reject::Rejection> {
    let body_: Vec<_> = body_.collect();
    let mut body: Cow<str> = "".into();
    for (k, v) in url::form_urlencoded::parse(&body_) {
        if k == "text" {
            body = v
        }
    }

    debug!("{:?}", body);
    let mut f = if cfg!(feature="docker") {
        std::fs::File::create(name.as_str()).unwrap()
    } else {
        let path = std::path::Path::new(name.as_str())
            .join("src")
            .join("main.rs");
        std::fs::create_dir_all(path.parent().unwrap()).unwrap();
        debug!("path = {:?}", path);
        std::fs::File::create(&path).unwrap()
    };
    f.write_all(body.as_bytes()).unwrap();

    (RunFuture { file: Some(name) })
        .and_then(|x| codenano::cadnano::to_cadnano(x.out.as_bytes()))
        .map_err(|e| warp::reject::custom(e))
}
*/

fn export_json(body_: warp::filters::body::FullBody) -> String {
    let body_: Vec<_> = body_.collect();
    let mut body: Cow<str> = "".into();
    for (k, v) in url::form_urlencoded::parse(&body_) {
        if k == "text" {
            body = v
        }
    }
    println!("body_ {}", body_.len());
    println!("in export_json");

    body.to_string()
}
