#[cfg(all(target_os = "linux", feature = "notify"))]
pub struct Notify {
    path: Option<std::path::PathBuf>,
    descr: Option<inotify::WatchDescriptor>,
    notify: inotify::Inotify,
    stream: inotify::EventStream<Vec<u8>>,
    rx: tokio::sync::mpsc::UnboundedReceiver<String>,
}

#[cfg(not(all(target_os = "linux", feature = "notify")))]
pub struct Notify {
    file: Option<std::path::PathBuf>,
    last_modification: std::time::SystemTime,
    interval: tokio::timer::Interval,
    rx: tokio::sync::mpsc::UnboundedReceiver<String>,
}

impl Notify {
    #[cfg(all(target_os = "linux", feature = "notify"))]
    pub fn new(
        file: Option<std::path::PathBuf>,
        rx: tokio::sync::mpsc::UnboundedReceiver<String>,
    ) -> Self {
        let mut notify = inotify::Inotify::init().unwrap();
        let descr = if let Some(ref file) = file {
            Some(notify.add_watch(file, inotify::WatchMask::MODIFY).unwrap())
        } else {
            None
        };
        Notify {
            path: file,
            descr,
            stream: notify.event_stream(vec![0; 1024]),
            notify,
            rx,
        }
    }
    #[cfg(not(all(target_os = "linux", feature = "notify")))]
    pub fn new(
        file: Option<std::path::PathBuf>,
        rx: tokio::sync::mpsc::UnboundedReceiver<String>,
    ) -> Self {
        let last_modification = if let Some(ref file) = file {
            std::fs::metadata(file).unwrap().modified().unwrap()
        } else {
            std::time::SystemTime::UNIX_EPOCH
        };
        Notify {
            file,
            last_modification,
            interval: tokio::timer::Interval::new_interval(std::time::Duration::new(0, 1_000_000)),
            rx,
        }
    }
}

impl futures::Stream for Notify {
    type Item = Option<std::path::PathBuf>;
    type Error = ();
    #[cfg(all(target_os = "linux", feature = "notify"))]
    fn poll(&mut self) -> futures::Poll<Option<Self::Item>, Self::Error> {
        use futures::Async;
        match self.stream.poll() {
            Ok(Async::Ready(Some(_))) => {
                return Ok(futures::Async::Ready(self.path.clone().map(Some)))
            }
            Ok(Async::NotReady) => (),
            Ok(Async::Ready(None)) => return Ok(Async::Ready(None)),
            Err(_) => return Err(()),
        };
        // Checking for updates
        match self.rx.poll() {
            Ok(futures::Async::NotReady) => Ok(futures::Async::NotReady),
            Ok(futures::Async::Ready(Some(new_path))) => {
                let path = std::path::Path::new(&new_path);
                if let Some(d) = self.descr.take() {
                    self.notify.rm_watch(d).unwrap();
                }
                self.descr = Some(
                    self.notify
                        .add_watch(path, inotify::WatchMask::MODIFY)
                        .unwrap(),
                );
                let p = path.to_path_buf();
                self.path = Some(p.clone());
                Ok(futures::Async::Ready(Some(Some(p))))
            }
            Ok(futures::Async::Ready(None)) => Ok(futures::Async::Ready(None)),
            Err(_) => Err(()),
        }
    }
    #[cfg(not(all(target_os = "linux", feature = "notify")))]
    fn poll(&mut self) -> futures::Poll<Option<Self::Item>, Self::Error> {
        let mut update_needed = false;
        match self.rx.poll() {
            Ok(futures::Async::Ready(Some(new_path))) => {
                let path = std::path::Path::new(&new_path);
                self.file = Some(path.to_path_buf());
                update_needed = true;
            }
            _ => (),
        }
        if let Ok(futures::Async::Ready(_)) = self.interval.poll() {
            update_needed = true
        }
        debug!("update needed: {:?}", update_needed);
        if update_needed {
            if let Some(ref file) = self.file {
                if let Ok(time) = std::fs::metadata(&file).and_then(|x| x.modified()) {
                    if time > self.last_modification {
                        self.last_modification = time;
                        return Ok(futures::Async::Ready(Some(Some(file.to_path_buf()))));
                    }
                } else {
                    return Err(());
                }
            }
            return Ok(futures::Async::Ready(Some(None)));
        }
        Ok(futures::Async::NotReady)
    }
}
